<?php

namespace Drupal\simple_survey\Service;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Datetime\DateFormatter;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use SplTempFileObject;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Url;
use SplFileObject;

/**
 * Survey Response convert and export service.
 *
 * @todo Add multivalue question value validation.
 */
class ResponseExportService implements ContainerInjectionInterface {

  /**
   * The module name.
   *
   * @var string
   */
  private $moduleName = 'simple_survey';

  /**
   * Drupal config service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  private $configFactory;

  /**
   * Drupal logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Drupal messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The Drupal Time Service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * Date formatter service.
   *
   * @var \Drupal\Core\Datetime\DateFormatter
   */
  protected $dateFormatter;

  /**
   * The Entity Type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The current survey.
   *
   * @var \Drupal\simple_survey\Entity\SimpleSurvey
   */
  public $survey;

  /**
   * PHP File Object.
   *
   * @var \SplTempFileObject
   */
  public $fileObject;

  /**
   * The export file name.
   *
   * @var string
   */
  public $filename;

  /**
   * Header keys.
   *
   * @var array
   */
  public $headerColumnKeys = [
    'drupal_id',
    'cookie_uuid',
    'site_path',
    'langcode',
  ];

  /**
   * Service config and DI.
   *
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   Drupal Messenger service.
   * @param \Drupal\Core\Datetime\DateFormatter $dateFormatter
   *   Date formatter service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $loggerFactory
   *   Drupal logger service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   Drupal config factory service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity Type Manager service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   Time service.
   */
  public function __construct(
    MessengerInterface $messenger,
    DateFormatter $dateFormatter,
    LoggerChannelFactoryInterface $loggerFactory,
    ConfigFactoryInterface $configFactory,
    EntityTypeManagerInterface $entity_type_manager,
    TimeInterface $time
  ) {
    $this->messenger = $messenger;
    $this->configFactory = $configFactory;
    $this->dateFormatter = $dateFormatter;
    $this->entityTypeManager = $entity_type_manager;
    $this->logger = $loggerFactory->get($this->moduleName);
    $this->time = $time;

    $this->simpleSurveyConfig = $this->configFactory->getEditable($this->moduleName . '.settings');
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('messenger'),
      $container->get('date.formatter'),
      $container->get('logger.factory'),
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('datetime.time')
    );
  }

  /**
   * Builds the csv header row and sets column keys.
   *
   * @return array
   *   Returns an array of header data.
   */
  public static function createHeaderRow(string $survey) {
    $db = \Drupal::service('database');

    $header = [
      'columns' => [],
      'keys' => [
        'drupal_id',
        'cookie_uuid',
        'site_path',
        'langcode',
      ],
      'validation' => [],
    ];

    $header['columns'] = [
      preg_replace('/[^A-Za-z0-9_]/', '', strtolower(str_replace(' ', '_', 'Drupal ID'))),
      preg_replace('/[^A-Za-z0-9_]/', '', strtolower(str_replace(' ', '_', 'User'))),
      preg_replace('/[^A-Za-z0-9_]/', '', strtolower(str_replace(' ', '_', 'Page Submitted On'))),
      preg_replace('/[^A-Za-z0-9_]/', '', strtolower(str_replace(' ', '_', 'Language'))),
    ];

    /** @var \Drupal\Core\Database\Query\SelectInterface $query */
    $query = $db->select('survey_question', 'sq');

    $query->condition('sq.survey', $survey);
    $query->fields('sq', ['id', 'question__value', 'question_type']);
    $result = $query->execute();
    $questions = $result->fetchAll();

    foreach ($questions as $question) {
      if ($question->question_type == 'normal') {
        $sanitizeQuestion = strip_tags($question->question__value);
        $sanitizeQuestion = str_replace('&nbsp;', ' ', $sanitizeQuestion);
        $sanitizeQuestion = str_replace(["\r", "\n"], '', $sanitizeQuestion);
        $header['columns'][] = preg_replace('/[^A-Za-z0-9_]/', '', strtolower(str_replace(' ', '_', $sanitizeQuestion)));
        $header['keys'][] = $question->id;

        $header['validation'][$question->id] = [];
      }
    }

    $header['columns'][] = preg_replace('/[^A-Za-z0-9_]/', '', strtolower(str_replace(' ', '_', 'Submission was created')));
    $header['keys'][] = 'created';

    return $header;
  }

  /**
   * Gets number of responses for current survey.
   *
   * @return int
   *   The number of responses.
   */
  public static function getResponseCount($survey) {
    $db = \Drupal::service('database');
    /** @var \Drupal\Core\Database\Query\SelectInterface $query */
    $query = $db->select('survey_response', 'sr');

    if ($survey) {
      $query->condition('sr.survey', $survey);
    }

    return $query
      ->countQuery()
      ->execute()->fetchField();
  }

  /**
   * Builds a response data rray.
   */
  public static function createCsvRow($keys, $validation, $surveyResponse) {
    $db = \Drupal::service('database');
    $outputResponse = [];

    foreach ($keys as $key) {
      $outputResponse[$key] = NULL;
    }

    $outputResponse['drupal_id'] = $surveyResponse->id;
    $outputResponse['cookie_uuid'] = $surveyResponse->cookie_uuid;
    $outputResponse['site_path'] = $surveyResponse->site_path;
    $outputResponse['langcode'] = $surveyResponse->langcode;
    $outputResponse['created'] = \Drupal::service('date.formatter')->format($surveyResponse->created, 'custom', 'Y-m-d H:i:s');

    $query = $db->select('survey_answer', 'sa');

    $survey_answers = $query
      ->condition('parent_uuid', $surveyResponse->uuid)
      ->fields('sa')
      ->execute()
      ->fetchAll();

    foreach ($survey_answers as $survey_answer) {
      $answer = $survey_answer->answer;
      $question_id = $survey_answer->question_id;

      $outputResponse[$question_id] = str_replace(['\r', '\n'], ' ', $answer);
    }

    return $outputResponse;
  }

  /**
   * Get the file object size.
   *
   * @return int
   */
  public function getFileSize() {
    return $this->fileObject->getSize();
  }

  public function outputToBrowser() {
    $this->fileObject->rewind();
    $this->fileObject->fpassthru();
  }

  /**
   * Batch process method for survey responses.
   *
   * @param string $filename
   *   The filename.
   * @param string $survey
   *   The survey machine name.
   * @param array|object $context
   *   The batch process context.
   */
  public static function processData(string $filename, string $survey, &$context) {
    $batch_size = 500;
    $db = \Drupal::service('database');
    $filepath = \Drupal::service('file_system')->realpath("temporary://")
      . '/' . $filename;

    if (empty($context['sandbox'])) {
      $file = new SplFileObject($filepath, 'w');
      $context['sandbox']['headers'] = self::createHeaderRow($survey);
      $context['sandbox']['max'] = self::getResponseCount($survey);
      $context['sandbox']['progress'] = 0;
      $context['results']['filepath'] = $filepath;
      $context['results']['filename'] = $filename;
      $context['results']['survey'] = $survey;
      $file->fputcsv($context['sandbox']['headers']['columns']);
    }
    else {
      $file = new SplFileObject($filepath, 'a');
    }

    $max = $context['sandbox']['progress'] + $batch_size;

    $surveyResponses = $db->select('survey_response', 'sr')
      ->condition('sr.survey', $survey)
      ->fields('sr')
      ->orderBy('sr.created', 'DESC')
      ->range($context['sandbox']['progress'], $batch_size)
      ->execute()
      ->fetchAll();

    if ($max > $context['sandbox']['max']) {
      $max = $context['sandbox']['max'];
    }

    $context['message'] = t('Adding rows @starting through @max of @total to @filename.<br>
    Current Size: @fileSize. | Memory used: @memUsage', [
      '@starting' => $context['sandbox']['progress'],
      '@max' => $max,
      '@total' => $context['sandbox']['max'],
      '@filename' => $filename,
      '@fileSize' => format_size($file->getSize()),
      '@memUsage' => format_size(memory_get_usage(TRUE)),
    ]);

    foreach ($surveyResponses as $surveyResponse) {
      $outputResponse = self::createCsvRow(
        $context['sandbox']['headers']['keys'],
        $context['sandbox']['headers']['validation'],
        $surveyResponse);

      $file->fputcsv($outputResponse);
      $context['sandbox']['progress']++;
    }

    if ($context['sandbox']['progress'] != $context['sandbox']['max']) {
      $context['finished'] = $context['sandbox']['progress'] / $context['sandbox']['max'];
    }
  }

  /**
   * CSV file creation atch process finished method.
   *
   * @param bool $success
   *   Boolean value that specifies batch success.
   * @param array $results
   *   Array that contains results from batch context.
   * @param array $operations
   *   Array that contains batch operations.
   */
  public static function finishedFileProcessing(bool $success, array $results, array $operations) {
    $status = Messenger::TYPE_STATUS;

    if ($success) {
      $message = t('CSV File created. This file will be available for 6 hours.
      <br>Download: <a href=":link" title=":filename">:filename</a>', [
        ':link' => Url::fromRoute('simple_survey.download_responses', [
          'simple_survey' => $results['survey'],
          'filename' => $results['filename'],
        ])->toString(),
        ':filename' => $results['filename'],
      ]);

      \Drupal::logger('simple_survey')->info($message);
    }
    else {
      $message = t('Failed to build CSV File, :filename.', [
        ':filename' => $results['filename'],
      ]);

      $status = Messenger::TYPE_ERROR;
      \Drupal::logger('simple_survey')->error($message);
    }

    \Drupal::messenger()->addMessage($message, $status);
  }

}
